package com.cxytiandi.cache_sdk;

import java.util.concurrent.TimeUnit;
import java.util.function.Function;
import org.springframework.beans.factory.annotation.Autowired;
import com.alicp.jetcache.Cache;
import com.alicp.jetcache.anno.SerialPolicy;
import com.alicp.jetcache.anno.support.ConfigProvider;
import com.alicp.jetcache.embedded.LinkedHashMapCacheBuilder;
import com.cxytiandi.cache_sdk.lock.DefaultLock;
import com.cxytiandi.cache_sdk.lock.Lock;
import com.cxytiandi.cache_sdk.ratelimit.DefaultRateLimit;
import com.cxytiandi.cache_sdk.ratelimit.RateLimit;
import com.cxytiandi.cache_sdk.utils.CacheKeyUtils;

import redis.clients.jedis.Jedis;
import redis.clients.util.Pool;

public class CachePlusImpl implements CachePlus {

	private Lock lock;
	
	private RateLimit rateLimit;
	
	@Autowired
	private Pool<Jedis> defaultPool;
	
	@Autowired
	protected ConfigProvider configProvider;
	
	Cache<String, Object> localCache = LinkedHashMapCacheBuilder.createLinkedHashMapCacheBuilder()
            .limit(100)
            .expireAfterWrite(200, TimeUnit.SECONDS)
            .buildCache();
	
	public CachePlusImpl() {
		this.lock = new DefaultLock();
		this.rateLimit = new DefaultRateLimit();
	}
	
	public CachePlusImpl(Lock lock, RateLimit rateLimit) {
		this.lock = lock;
		this.rateLimit = rateLimit;
	}

	public <V, K> V getCache(String name, K key, Closure<V, K> closure) {
		return this.doGetCache(name, key, closure, 0, TimeUnit.SECONDS);
	}

	@Override
	public <V, K> V getCache(String name, K key, Closure<V, K> closure, long expire, TimeUnit timeUnit) {
		return this.doGetCache(name, key, closure, expire, timeUnit);
	}
	
	@SuppressWarnings("unchecked")
	public <V, K> V doGetCache(String name, K key, Closure<V, K> closure, long expire, TimeUnit timeUnit) {
		// 转换缓存Key
		String jsonKey = CacheKeyUtils.convertKey(name, key);
		
		// 热点Key本地读取
		if (HotKey.hasHotKey(jsonKey)) {
			Object localValue = localCache.get(jsonKey);
			if (localValue != null) {
				return (V) localValue;
			}
		}
		Jedis jedis = defaultPool.getResource();
		// 获取Key对应的缓存数据
		Object cacheData = getCacheByRedis(jedis, jsonKey);
		if (cacheData != null) {
			return (V) cacheData;
		}
		// 没有缓存数据，调用加载逻辑进行数据缓存
		V result = null;
		try {
			rateLimit.acquire();
			lock.lock(jsonKey);
			cacheData = getCacheByRedis(jedis, jsonKey);
			if (cacheData != null) {
				return (V) cacheData;
			}
			result = closure.execute(key);
			// 编码
			Function<Object, byte[]> f = configProvider.parseValueEncoder(SerialPolicy.JAVA);
			byte[] data = f.apply(result);
			// 缓存
			if (expire > 0) { 
				jedis.set(jsonKey.getBytes(), data, "NX".getBytes(), "EX".getBytes(), timeUnit.toSeconds(expire));
			} else {
				jedis.set(jsonKey.getBytes(), data);
			}
		} catch (Exception e) {
			e.printStackTrace();
		} finally {
			jedis.close();
			lock.unLock(jsonKey);
			rateLimit.release();
		}
		return result;
	}
	
	private Object getCacheByRedis(Jedis jedis,  String jsonKey) {
		Object cacheData = null;
		byte[] array = jedis.get(jsonKey.getBytes());
		if (array != null) {
			try {
				// 获取到了就解码
				Function<byte[], Object> f = configProvider.parseValueDecoder(SerialPolicy.JAVA);
				cacheData =  f.apply(array);
				// 热点Key加入本地缓存
				if (HotKey.hasHotKey(jsonKey)) {
					localCache.put(jsonKey, cacheData);
				}
			} catch (Exception e) {
				e.printStackTrace();
			} finally {
				jedis.close();
			}
		}
		return cacheData;
	}

}
